<?php

namespace App\Model;

use Nette;
use Nette\Database\Table\ActiveRow;
use Nette\Http\FileUpload;

/**
 * Class for handling activities related to assignments
 * @package App\Model
 */
class AssignmentManager {
    use Nette\SmartObject;

    private Nette\Database\Explorer $database;
    private UserManager $userManager;

    public function __construct(Nette\Database\Explorer $database, UserManager $userManager) {
        $this->database = $database;
        $this->userManager = $userManager;
    }

    /**
     * Function returns assignment of given identifier
     * @param int $id assignment identifier
     * @return ActiveRow|null
     */
    public function getAssignment(int $id):ActiveRow|null {
        return $this->database->table("ASSIGNMENT")->get($id);
    }

    /**
     * Function returns contact record of given assignment record
     * @param ActiveRow $assignment
     * @return ActiveRow|null
     */
    public function getCreator(ActiveRow|null $assignment):ActiveRow|null {
        if($assignment == null) return null;

        return $assignment->ref('CONTACT', 'contact_id');
    }

    /**
     * Function returns contact of user as a mentor of given assignment record
     * @param ActiveRow $assignment
     */
    public function getMentor(ActiveRow|null $assignment):ActiveRow|null
    {
        if($assignment->user != null) {
            return $assignment->ref('USER', 'user_id')->ref('CONTACT');
        }

        return null;
    }

    /**
     * Function return user as mentor of given assignment
     * @param ActiveRow|null $assignment
     * @return ActiveRow|null
     */
    public function getMentorUser(ActiveRow|null $assignment) {
        if($assignment == null) return null;

        return $assignment->ref('USER');
    }

    /**
     * Assign mentor to assignment
     * @param ActiveRow|null $assignment assignment record
     * @param $mentor_id mentor user identifier
     * @return bool
     */
    public function assignMentor(ActiveRow|null $assignment, $mentor_id):bool {
        if($assignment->user != null) {
            return false;
        }

        if(!$this->userManager->isInRole($mentor_id, RoleManager::$ROLE_MENTOR)) {
            return false;
        }

        $this->database->query('UPDATE ASSIGNMENT set user_id = ? where id = ?', $mentor_id, $assignment->id);
        return true;

    }

    /**
     * Function to check if assignment is assigned
     * @param $id assignment identifier
     * @return bool
     */
    public function isAssigned($id):bool {
        $count = $this->database->table('ASSIGNMENT')->get($id)->related('PROJECT', 'assignment_id')->count();
        return $count > 0;
    }

    /**
     * Function return related project of given assignment
     * @param $id assignment identifier
     * @return ActiveRow
     */
    public function getAssignedProject($id):ActiveRow {
        return $this->database->table('ASSIGNMENT')->get($id)->related('PROJECT')->fetch();
    }

    /**
     * Function returns not assigned assignments
     * @return Nette\Database\Table\Selection
     */
    public function getFreeAssignments() {
        return $this->database->table('ASSIGNMENT')->having("COUNT(:PROJECT.id) = 0")->group('ASSIGNMENT.id');
    }

    /**
     * Create new project from assignment interest
     * @param ActiveRow $interest assignment interest record
     * @return bool
     */
    public function assignTeamByInterest(ActiveRow $interest):bool {
        $this->database->table('PROJECT')->insert([
            'assignment_id' => $interest->assignment_id,
            'team_id' => $interest->team_id,
        ]);

        $this->database->table('ASSIGNMENT_INTEREST')->where('assignment_id', $interest->assignment_id)->update(['state' => 1]);
        $this->database->table('ASSIGNMENT_INTEREST')->where('team_id', $interest->team_id)->update(['state' => 2]);
        $this->database->table('ASSIGNMENT_INTEREST')->where('team_id', $interest->team_id)->where('assignment_id', $interest->assignment_id)->update(['state' => 4]);
        return true;
    }

    /**
     * Function return functions that has assigned mentors
     * @param $mentorId
     * @return Nette\Database\Table\GroupedSelection
     */
    public function getMentoredAssignments($mentorId):Nette\Database\Table\GroupedSelection {
        return $this->database->table('USER')->get($mentorId)->related('ASSIGNMENT');
    }

    /**
     * Function returns mentored assignments that is not assigned by any team and do not relate to any assignment interest
     * @param $mentorId
     * @return Nette\Database\Table\GroupedSelection
     */
    public function getMentoredAssignmentsWithoutTeamAndInterest($mentorId) {
        return $this->database->table('USER')->get($mentorId)->related('ASSIGNMENT');
    }

    /**
     * Function returns record of assignment file related to given assignment record
     * @param ActiveRow $assignment
     * @return Nette\Database\Table\GroupedSelection
     */
    public function getAssignmentFiles(ActiveRow $assignment) {
        return $assignment->related('ASSIGNMENT_FILE', 'assignment_id');
    }

    /**
     * Function to get record of assignment file
     * @param $fileId assignment file identifier
     * @return ActiveRow
     */
    public function getAssignmentFile($fileId):ActiveRow {
        return $this->database->table('ASSIGNMENT_FILE')->get($fileId);
    }

    /**
     * Function returns selection of assignment that has not assigned mentor
     * @return Nette\Database\Table\Selection
     */
    public function getAssignmentsWithoutMentor() {
        return $this->database->table('ASSIGNMENT')->where('user_id IS NULL');
    }

    /**
     * Function returns selection of assignment that has assigned mentor
     * @return Nette\Database\Table\Selection
     */
    public function getAssignmentsWithMentor() {
        return $this->database->table('ASSIGNMENT')->where('user_id IS NOT NULL');
    }

    /**
     * Function returns profile badges of given assignment
     * @param ActiveRow $assignment
     * @return Nette\Database\Table\GroupedSelection
     */
    public function getAssignmentProfiles(ActiveRow $assignment) {
        return $assignment->related('ASSIGNMENT_TAG')->select('TAG.id AS id, TAG.name AS name');
    }

    /**
     * Function to create new assignment from given array
     * @param $values
     * @return ActiveRow|null
     */
    public function createAssignment($values):ActiveRow|null {
        $this->database->beginTransaction();
        $contact = $this->database->table('CONTACT')->insert([
            'firstname' => $values['creatorFirstname'],
            'lastname' => $values['creatorLastname'],
            'email' => $values['creatorEmail'],
            'phone' => $values['creatorPhone'],
            'info' => $values['creatorInfo'],
        ]);

        $assignment = $this->database->table('ASSIGNMENT')->insert([
            'name' => $values['name'],
            'description' => $values['description'],
            'team_size' => $values['teamSize'],
            'period' => $values['period'],
            'user_id' => $values['mentor'] == 0 ? null : $values['mentor'],
            'contact_id' => $contact->id,
            'duration' => $values['duration'],
            'hidden' => $values['visibility'],
        ]);

        $this->database->table("ASSIGNMENT_TAG")->where("assignment", $assignment->id)->delete();

        $profiles = array();
        foreach ($values['profile'] as $profile) {
            $profiles[] = [
                "assignment" => $assignment->id,
                "tag_id" => $profile,
            ];
        }

        if(count($profiles) > 0) {
            $this->database->table('ASSIGNMENT_TAG')->insert($profiles);
        }

        if(isset($values['pdfFile'])) {
            $this->uploadAssignmentFileInsert($values['pdfFile'], $assignment->id, $values['userId']);
        }

        $this->database->commit();

        return $assignment;
    }

    /**
     * Function edits assignment with given parameter array
     * @param $values
     * @return ActiveRow|null
     */
    public function editAssignment($values) {

        $this->database->beginTransaction();
        $assignment = $this->database->table('ASSIGNMENT')->get($values['assignmentId']);

        $assignment->contact->update([
            'firstname' => $values['creatorFirstname'],
            'lastname' => $values['creatorLastname'],
            'email' => $values['creatorEmail'],
            'phone' => $values['creatorPhone'],
            'info' => $values['creatorInfo'],
        ]);

        $assignment->update([
            'name' => $values['name'],
            'description' => $values['description'],
            'team_size' => $values['teamSize'],
            'period' => $values['period'],
            'user_id' => $values['mentor'] == 0 ? null : $values['mentor'],
            'duration' => $values['duration'],
            'hidden' => $values['visibility'],
        ]);

        $this->database->table("ASSIGNMENT_TAG")->where("assignment", $assignment->id)->delete();

        $profiles = array();
        foreach ($values['profile'] as $profile) {
            $profiles[] = [
                "assignment" => $assignment->id,
                "tag_id" => $profile,
            ];
        }

        if(count($profiles) > 0) {
            $this->database->table('ASSIGNMENT_TAG')->insert($profiles);
        }

        if(isset($values['pdfFile'])) {
            $this->uploadAssignmentFileUpdate($values['pdfFile'], $assignment->id, $values['userId']);
        }

        $this->database->commit();

        return $assignment;
    }

    /**
     * Function to upload assignment file
     * @param FileUpload $file
     * @param $assignmentId
     * @param $userId
     */
    public function uploadAssignmentFileInsert(FileUpload $file, $assignmentId, $userId) {

        if(!$file->isOk()) return;

        if($file->size > 0) {
            $this->database->table('ASSIGNMENT_FILE')->insert([
                'file' => $file->getContents(),
                'assignment_id' =>  $assignmentId,
                'user_id' => $userId,
                'filesize' => $file->size,
                'filename' => $file->name,
            ]);

        }
    }

    /**
     * Function to upload assignment file
     * @param FileUpload $file
     * @param $assignmentId
     * @param $userId
     */
    public function uploadAssignmentFileUpdate(FileUpload $file, $assignmentId, $userId) {

        if(!$file->isOk()) return;

        if($file->size > 0) {
            $this->database->table('ASSIGNMENT_FILE')->where('assignment_id', $assignmentId)->update([
                'file' => $file->getContents(),
                'user_id' => $userId,
                'filesize' => $file->size,
                'filename' => $file->name,
            ]);
        }
    }
}