<?php

namespace App\Model;

use Nette;
use Nette\Utils\DateTime;
use Nette\Database\Table\ActiveRow;
use Nette\Database\Table\GroupedSelection;

/**
 * Class for handling activities related to assignment interests
 * @package App\Model
 */
class InterestManager {

    use Nette\SmartObject;

    private Nette\Database\Explorer $database;

    public function __construct(Nette\Database\Explorer $database) {
        $this->database = $database;
    }

    /**
     * Function to create assignment interest record from given parameters
     * @param $assignmentId assignment identifier
     * @param $teamId team identifier
     * @param DateTime $datetime project start date
     * @param int $weight weight of interest
     * @return ActiveRow|null
     */
    public function createInterest($assignmentId, $teamId, DateTime $datetime, int $weight): ActiveRow|null {

        return $this->database->table('ASSIGNMENT_INTEREST')->insert([
            'assignment_id' => $assignmentId,
            'team_id' => $teamId,
            'created_time' => $datetime,
            'weight' => $weight,
        ]);
    }

    /**
     * Function to check if assignment has interest from given team
     * @param $teamId team identifier
     * @param $assignmentId assignment identifier
     * @return bool
     */
    public function hasTeamInterest($teamId, $assignmentId):bool {
        $count = $this->database->table('ASSIGNMENT_INTEREST')->where('team_id = ?', $teamId)->where('assignment_id = ?', $assignmentId)->count();
        return $count > 0;
    }

    /**
     * Function returns to get related team interests
     * @param $teamId
     * @return Nette\Database\Table\Selection
     */
    public function getTeamInterests($teamId) {
        return $this->database->table('ASSIGNMENT_INTEREST')->where('team_id = ?', $teamId)->order('state DESC, weight DESC');
    }

    /**
     * Function returns interests of given team
     * @param ActiveRow $team team record
     * @return GroupedSelection
     */
    public function getInterestsByTeam(Nette\Database\Table\ActiveRow $team) {
        return $team->related('ASSIGNMENT_INTEREST.team_id')->order('state DESC, weight DESC');
    }

    /**
     * Function returns interests of assignments that is mentored by given user
     * @param $mentor_id user identifier
     * @return Nette\Database\Table\Selection
     */
    public function getInterestsByMentor($mentor_id) {
        return $this->database->table('ASSIGNMENT_INTEREST')->where('state', 3)->where('ASSIGNMENT.user_id', $mentor_id);
    }

    /**
     * Function returns interest of given assignment and team
     * @param $assignmentId assignment identifier
     * @param $teamId team identifier
     * @return ActiveRow|null
     */
    public function getInterest($assignmentId, $teamId) {
        return $this->database->table('ASSIGNMENT_INTEREST')->get(['team_id' => $teamId, 'assignment_id' => $assignmentId]);
    }

    /**
     *
     * @param ActiveRow $team
     * @return GroupedSelection
     */
    public function getActiveInterests(ActiveRow $team):GroupedSelection {
        return $team->related('ASSIGNMENT_INTEREST')->where('state', 3)->limit(5);
    }

    /**
     * Mentor interest evaliation
     * @param $teamId team identifier
     * @param $assignmentId assignment identifier
     * @param $evaluation evaluatoin identifier
     * @return int
     */
    public function evaluate($teamId, $assignmentId, $evaluation):int {
        if($evaluation >= 0 && $evaluation <= 2) {
            return $this->database->table('ASSIGNMENT_INTEREST')->where('assignment_id', $assignmentId)->where('team_id', $teamId)->update(['mentor_evaluation' => $evaluation]);
        } else {
            return -1;
        }
    }

}