<?php

namespace App\Presenters;

use App\Components\StudentAddForm;
use App\Components\StudentImportForm;
use App\Components\TeamCreationForm;
use App\Model\MailManager;
use App\Model\RoleManager;
use App\Model\StudentManager;
use App\Model\TeamManager;
use App\Model\UserManager;
use App\Model\Utils;
use Nette\Application\UI\Form;
use Nette\Utils\ArrayHash;

/**
 * Presenter class to handle requests to page with student management
 */
class StudentManagementPresenter extends BasePresenter {

    protected StudentManager $studentManager;
    protected StudentImportForm $studentImportForm;
    protected StudentAddForm $studentAddForm;
    protected TeamCreationForm $teamCreationForm;
    protected UserManager $userManager;
    protected MailManager $mailManager;

    public function __construct(StudentManager $studentManager,
                                StudentImportForm $studentImportForm,
                                StudentAddForm $studentAddForm,
                                TeamCreationForm $teamCreationForm,
                                UserManager $userManager,
                                MailManager $mailManager) {
        parent::__construct();

        $this->studentManager = $studentManager;
        $this->studentImportForm = $studentImportForm;
        $this->studentAddForm = $studentAddForm;
        $this->teamCreationForm = $teamCreationForm;
        $this->userManager = $userManager;
        $this->mailManager = $mailManager;
    }

    public function startup() {
        parent::startup();

        if($this->getAction() == "createTeam") {
            $id = intval($this->getParameter('id'));

            $student = $this->studentManager->getStudent($id);
            if(!$student) {
                $this->flashMessage('Chyba;Student neexistuje', 'bg-danger');
                $this->redirect('StudentManagement:');
            }

            if(!empty($student->team)) {
                $this->flashMessage('Chyba;Student již je součástí jiného týmu.', 'bg-danger');
                $this->redirect('StudentManagement:');
            }
        }

    }

    public function renderDefault() {
        $this->template->students = $this->studentManager->getAllStudents()->select('STUDENT.id, firstname, lastname, personalID, email, team_id, :TEAM.id AS leader');

    }

    public function renderCreateTeam(int $id) {
    }

    public function createComponentStudentImportForm():Form {

        $form = $this->studentImportForm->create();
        $form->onSuccess[] = function() {
            $this->flashMessage('Úspěch;Import studentů proběhl úspěšně', 'bg-success');
            $this->redirect('StudentManagement:');
        };

        return $form;
    }

    public function createComponentStudentAddForm():Form {
        $form = $this->studentAddForm->create();

        $form->onSuccess[] = function() {
            $this->flashMessage('Úspěch;Vytvoření studenta proběhlo úspěšně', 'bg-success');
            $this->redirect('StudentManagement:');
        };

        return $form;
    }

    public function createComponentTeamCreationForm():Form {
        $leaderId = intval($this->getParameter('id'));
        $form = $this->teamCreationForm->create($leaderId);
        $form->onSuccess[] = function(Form $form, ArrayHash $values) use ($leaderId) {

            $team = $this->teamManager->createTeam($values['leaderId'], $values['teamName']);
            $leader = $this->studentManager->getStudent($values['leaderId']);
            $password = Utils::getGeneratedPassword();

            if ($this->configManager->getConfig()['testing']['enabled']) {
                if ($this->configManager->getConfig()['testing']['testPassword']) {
                    $password = $this->configManager->getConfig()['testing']['testPasswordValue'];
                }
            }

            $this->userManager->createStudentUser($leader->login, $password, $leader->id, $leader->email);
            $this->mailManager->sendLeaderCreated($leader->email, $leader->login, $password);
            $this->flashMessage('Úspěch;Vytvoření týmu problěhlo úspěšně', 'bg-success');
            $teamId = $this->teamManager->getTeamIdRelByStudent(intval($leaderId));
            $this->loggerManager->get('default')->notice('Team has been created', [
                'user_id' => $this->getUser()->getIdentity()->getData(),
                'leader_id' => $leaderId,
                'leader_login' => $leader->login,
                'team_id' => $team->id,
                'team_name' => $team->name,
            ]);
            $this->redirect('Team:show', $teamId);
        };

        return $form;
    }

    public function handleStudentDelete(int $id) {
        if($this->studentManager->deleteStudent($id)) {
            $this->flashMessage('Úspěch;Odstranění studenta proběhlo úspěšně.', 'bg-success');
            $this->loggerManager->get('default')->info('Student has been removed', [
                'user_id' => $this->getUser()->getIdentity()->getData(),
                'student_id' => $id,
            ]);
        } else {
            $this->flashMessage('Chyba;Při odstraňování studenta se vyskytla chyba.', 'bg-danger');
        }
    }

    public function canAccessPage(): bool {
        if(!$this->user->isLoggedIn()) {
            return false;
        } else if($this->user->isInRole(RoleManager::$ROLE_GARANT)) {
            return true;
        }
        return false;
    }

}