<?php

namespace App\Components;

use App\Model\StudentManager;
use App\Model\TeamManager;
use App\Model\TagManager;
use App\Model\TeamRoleManager;
use App\Model\UserManager;
use App\Model\Utils;
use Nette\Application\AbortException;
use Nette\Application\UI\Form;
use Nette\Database\Explorer;
use Nette\Security\AuthenticationException;
use Nette\Security\User;
use Nette\Utils\ArrayHash;

/**
 * Form component class for assigning team roles to team members
 * @package App\Components
 */
class StudentAdvancesForm extends Form {
    protected TeamManager $teamManager;
    protected TeamRoleManager $teamRoleManager;


    public function __construct(TeamManager $teamManager, TeamRoleManager $teamRoleManager) {
        parent::__construct();
        $this->teamManager = $teamManager;
        $this->teamRoleManager = $teamRoleManager;
    }

    /**
     * Factory function for creating team member assigning roles form
     * @param int $teamId students team identifier
     * @return Form resulted Form instance
     * @throws \Exception
     */
    public function create(int $teamId = -1) {
        $defaults = [];
        $form = new Form();
        $renderer = $form->getRenderer();
        $renderer->wrappers['controls']['container'] = 'div class="m-2"';
        $renderer->wrappers['pair']['container'] = 'div class="form-group"';
        if ($teamId < 0) {
            throw new \Exception("Cannot create form with invalid id");
        }
        $teamMembers = $this->teamManager->getTeamMembers($teamId);
        $team = $this->teamManager->getTeam($teamId);
        $teamLeaderId = $team["student_id"];
        $values = $this->teamRoleManager->getAllApplicableRolesForSelect();
        $i = 0;
        foreach ($teamMembers as $id => $student) {
            $leaderString = ((int)$student["id"] === (int)$teamLeaderId ? " - VEDOUCÍ" : "");

            $roles = $this->teamRoleManager->getAllRolesForStudentForSelect($id);

            $form->addHidden("member_id" . $i, $id)->setHtmlAttribute("class", "member")->addRule($form::EQUAL, "Špatné hiddenId", $id);
            $studentInfo = Utils::transformStudentDataToString($student["firstname"],$student["lastname"],$student["personalID"],$leaderString);
            $form->addMultiSelect("members" . $i, $studentInfo, $values)
                ->addRule($form::MAX_LENGTH, 'Můžou být vybrány maximálně 3 předvolby', 3)
                ->setHtmlAttribute('class="selectpicker form-control" multiple data-live-search="true" title="Vyberte roli"')
                ->setHtmlAttribute("multiple=multiple")
                ->setHtmlAttribute("style= width:100%")->setDefaultValue(array_flip($roles));
            $i++;
        }


        $form->addHidden("teamId", $teamId)->addRule($form::EQUAL, "Špatné id teamu", $teamId);
        $form->addHidden("teamSize", count($teamMembers))->addRule($form::EQUAL, "Špatná velikost týmu", count($teamMembers));
        $form->addSubmit('save', 'Potvrdit změny')
            ->setHtmlAttribute('class', 'btn btn-primary')
            ->setHtmlId("frm-studentAdvancesForm-submit");
        $form->addButton('cancel', 'Zrušit')
            ->setHtmlAttribute('class', 'btn btn-link')
            ->setHtmlId("frm-studentAdvancesForm-cancel");

        //$form->onValidate[] = [$this, 'validateForm'];
        $form->onSuccess[] = [$this, 'saveForm'];

        return $form;
    }

    /**
     * Function that is triggered by a successful form submission
     * @param Form $form
     * @param ArrayHash $values
     */
    public function saveForm(Form $form, ArrayHash $values) {

        try {
            foreach ($values as $key => $value) {
                if (str_contains($key, "members")) {
                    $studCount = explode('s', $key)[1];

                    $id = $values['member_id' . $studCount];
                    $this->teamRoleManager->updateStudentRoles($id, $value);
                }

            }
        } catch (\Exception $e) {

        }
    }
}